@@ -1,6 +1,7 @@
 from letmedoit import config
 from packaging import version
-import platform, subprocess, os, pydoc, webbrowser, re, socket, wcwidth, unicodedata, traceback
+from bs4 import BeautifulSoup
+import platform, shutil, subprocess, os, pydoc, webbrowser, re, socket, wcwidth, unicodedata, traceback, html2text
 import datetime, requests, netifaces, textwrap, json, geocoder, base64, getpass, pendulum, pkg_resources
 import pygments
 from pygments.lexers.python import PythonLexer
@@ -9,29 +10,94 @@ from prompt_toolkit.styles.pygments import style_from_pygments_cls
 from prompt_toolkit import print_formatted_text
 from prompt_toolkit.formatted_text import PygmentsTokens
 from prompt_toolkit import prompt
-try:
-    import tiktoken
-    tiktokenImported = True
-except:
-    tiktokenImported = False
+import tiktoken
+import openai
 from openai import OpenAI
+from urllib.parse import quote
+from pathlib import Path
 from PIL import Image
+if not config.isTermux:
+    from autogen.retrieve_utils import TEXT_FORMATS
 
 
+def check_openai_errors(func):
+    def wrapper(*args, **kwargs):
+        try: 
+            return func(*args, **kwargs)
+        except openai.APIError as e:
+            print("Error: Issue on OpenAI side.")
+            print("Solution: Retry your request after a brief wait and contact us if the issue persists.")
+        except openai.APIConnectionError as e:
+            print("Error: Issue connecting to our services.")
+            print("Solution: Check your network settings, proxy configuration, SSL certificates, or firewall rules.")
+        except openai.APITimeoutError as e:
+            print("Error: Request timed out.")
+            print("Solution: Retry your request after a brief wait and contact us if the issue persists.")
+        except openai.AuthenticationError as e:
+            print("Error: Your API key or token was invalid, expired, or revoked.")
+            print("Solution: Check your API key or token and make sure it is correct and active. You may need to generate a new one from your account dashboard.")
+            HealthCheck.changeAPIkey()
+        except openai.BadRequestError as e:
+            print("Error: Your request was malformed or missing some required parameters, such as a token or an input.")
+            print("Solution: The error message should advise you on the specific error made. Check the [documentation](https://platform.openai.com/docs/api-reference/) for the specific API method you are calling and make sure you are sending valid and complete parameters. You may also need to check the encoding, format, or size of your request data.")
+        except openai.ConflictError as e:
+            print("Error: The resource was updated by another request.")
+            print("Solution: Try to update the resource again and ensure no other requests are trying to update it.")
+        except openai.InternalServerError as e:
+            print("Error: Issue on OpenAI servers.")
+            print("Solution: Retry your request after a brief wait and contact us if the issue persists. Check the [status page](https://status.openai.com).")
+        except openai.NotFoundError as e:
+            print("Error: Requested resource does not exist.")
+            print("Solution: Ensure you are the correct resource identifier.")
+        except openai.PermissionDeniedError as e:
+            print("Error: You don't have access to the requested resource.")
+            print("Solution: Ensure you are using the correct API key, organization ID, and resource ID.")
+        except openai.RateLimitError as e:
+            print("Error: You have hit your assigned rate limit.")
+            print("Solution: Pace your requests. Read more in OpenAI [Rate limit guide](https://platform.openai.com/docs/guides/rate-limits).")
+        except openai.UnprocessableEntityError as e:
+            print("Error: Unable to process the request despite the format being correct.")
+            print("Solution: Please try the request again.")
+        except:
+            print(traceback.format_exc())
+    return wrapper
+
 class SharedUtil:
 
     # token limit
     # reference: https://platform.openai.com/docs/models/gpt-4
     tokenLimits = {
-        #"gpt-3.5-turbo-instruct": 4097,
-        "gpt-3.5-turbo": 4097,
+        "gpt-4-turbo-preview": 128000, # Returns a maximum of 4,096 output tokens.
+        "gpt-4-0125-preview": 128000, # Returns a maximum of 4,096 output tokens.
+        "gpt-4-1106-preview": 128000, # Returns a maximum of 4,096 output tokens.
+        "gpt-3.5-turbo": 4096, # After 16Feb2024, should be adjusted to 16385, which Returns a maximum of 4,096 output tokens.
         "gpt-3.5-turbo-16k": 16385,
-        "gpt-4-1106-preview": 128000, # official 128,000; but "This model supports at most 4096 completion tokens"; set 8192 here to work with LetMeDoIt AI dynamic token feature
-        #"gpt-4-vision-preview": 128,000, # used in plugin "analyze images"
         "gpt-4": 8192,
         "gpt-4-32k": 32768,
     }
 
+    @staticmethod
+    @check_openai_errors
+    def checkCompletion():
+        SharedUtil.setAPIkey()
+        config.oai_client.chat.completions.create(
+            model="gpt-3.5-turbo",
+            messages=[{"role": "user", "content" : "hello"}],
+            n=1,
+            max_tokens=10,
+        )
+
+    @staticmethod
+    def setAPIkey():
+        # instantiate a client that can shared with plugins
+        os.environ["OPENAI_API_KEY"] = config.openaiApiKey
+        config.oai_client = OpenAI()
+        # set variable 'OAI_CONFIG_LIST' to work with pyautogen
+        oai_config_list = []
+        for model in SharedUtil.tokenLimits.keys():
+            oai_config_list.append({"model": model, "api_key": config.openaiApiKey})
+        os.environ["OAI_CONFIG_LIST"] = json.dumps(oai_config_list)
+
     @staticmethod
     def getPackageInstalledVersion(package):
         try:
@@ -55,6 +121,25 @@ class SharedUtil:
         installed_version = SharedUtil.getPackageInstalledVersion(package)
         return (latest_version > installed_version)
 
+    # handle document path dragged to the terminal
+    @staticmethod
+    def isExistingPath(docs_path):
+        docs_path = docs_path.strip()
+        search_replace = (
+            ("^'(.*?)'$", r"\1"),
+            ('^(File|Folder): "(.*?)"$', r"\2"),
+        )
+        for search, replace in search_replace:
+            docs_path = re.sub(search, replace, docs_path)
+        if "\\ " in docs_path or "\(" in docs_path:
+            search_replace = (
+                ("\\ ", " "),
+                ("\(", "("),
+            )
+            for search, replace in search_replace:
+                docs_path = docs_path.replace(search, replace)
+        return docs_path if os.path.exists(os.path.expanduser(docs_path)) else ""
+
     @staticmethod
     def getCurrentDateTime():
         current_datetime = datetime.datetime.now()
@@ -66,10 +151,49 @@ class SharedUtil:
         print(trace if config.developer else "Error encountered!")
         return trace
 
+    @staticmethod
     def addTimeStamp(content):
         time = re.sub("\.[^\.]+?$", "", str(datetime.datetime.now()))
         return f"{content}\n[Current time: {time}]"
 
+    @staticmethod
+    def downloadWebContent(url, timeout=60, folder="", ignoreKind=False):
+        config.print2("Downloading web content ...")
+        hasExt = re.search("\.([^\./]+?)$", url)
+        supported_documents = TEXT_FORMATS[:]
+        supported_documents.remove("org")
+
+        response = requests.get(url, timeout=timeout)
+        folder = folder if folder and os.path.isdir(folder) else os.path.join(config.letMeDoItAIFolder, "temp")
+        filename = quote(url, safe="")
+        def downloadBinary(filename=filename):
+            filename = os.path.join(folder, filename)
+            with open(filename, "wb") as fileObj:
+                fileObj.write(response.content)
+            return filename
+        def downloadHTML(filename=filename):
+            filename = os.path.join(folder, f"{filename}.html")
+            with open(filename, "w", encoding="utf-8") as fileObj:
+                fileObj.write(response.text)
+            return filename
+
+        try:
+            if ignoreKind:
+                filename = downloadBinary()
+                config.print3(f"Downloaded at: {filename}")
+                return ("any", filename)
+            elif hasExt and hasExt.group(1) in supported_documents:
+                return ("document", downloadBinary())
+            elif SharedUtil.is_valid_image_url(url):
+                return ("image", downloadBinary())
+            else:
+                # download content as text
+                # Save the content to a html file
+                return ("text", downloadHTML())
+        except:
+            SharedUtil.showErrors()
+            return ("", "")
+
     @staticmethod
     def is_valid_url(url):
         # Regular expression pattern for URL validation
@@ -85,7 +209,7 @@ class SharedUtil:
     @staticmethod
     def is_valid_image_url(url): 
         try: 
-            response = requests.head(url) 
+            response = requests.head(url, timeout=30)
             content_type = response.headers['content-type'] 
             if 'image' in content_type: 
                 return True 
@@ -114,6 +238,15 @@ class SharedUtil:
         os.path.splitext(os.path.basename(image_path))[1]
         return f"data:image/png;base64,{base64_image}"
 
+    def getWebText(url):
+        try:
+            # Download webpage content
+            response = requests.get(url, timeout=30)
+            # Parse the HTML content to extract text
+            soup = BeautifulSoup(response.text, 'html.parser')
+            return soup.get_text()
+        except:
+            return ""
 
     @staticmethod
     def transformText(text):
@@ -127,12 +260,16 @@ class SharedUtil:
         return style_from_pygments_cls(get_style_by_name(theme))
 
     @staticmethod
-    def showAndExecutePythonCode(code):
+    def displayPythonCode(code):
         if config.developer or config.codeDisplay:
             config.print("```python")
             tokens = list(pygments.lex(code, lexer=PythonLexer()))
             print_formatted_text(PygmentsTokens(tokens), style=SharedUtil.getPygmentsStyle())
             config.print("```")
+
+    @staticmethod
+    def showAndExecutePythonCode(code):
+        SharedUtil.displayPythonCode(code)
         config.stopSpinning()
         refinedCode = SharedUtil.fineTunePythonCode(code)
         information = SharedUtil.executePythonCode(refinedCode)
@@ -160,17 +297,14 @@ class SharedUtil:
 
     @staticmethod
     def getPythonFunctionResponse(code):
-        return str(config.pythonFunctionResponse) if config.pythonFunctionResponse is not None and type(config.pythonFunctionResponse) in (int, float, str, list, tuple, dict, set, bool) and not ("os.system(" in code) else ""
+        return str(config.pythonFunctionResponse) if config.pythonFunctionResponse is not None and (type(config.pythonFunctionResponse) in (int, float, str, list, tuple, dict, set, bool) or str(type(config.pythonFunctionResponse)).startswith("<class 'numpy.")) and not ("os.system(" in code) else ""
 
     @staticmethod
     def autoHealPythonCode(code, trace):
         for i in range(config.max_consecutive_auto_heal):
             userInput = f"Original python code:\n```\n{code}\n```\n\nTraceback:\n```\n{trace}\n```"
-            config.print(f"Auto-heal attempt {(i + 1)} ...")
-            config.print(f"running python code ...")
-            if config.developer:
-                print(f"```\n{code}\n```")
-            function_call_message, function_call_response = SharedUtil.getSingleFunctionResponse(userInput, config.heal_python_signature, "heal_python")
+            config.print3(f"Auto-correction attempt: {(i + 1)}")
+            function_call_message, function_call_response = SharedUtil.getSingleFunctionResponse(userInput, [config.chatGPTApiFunctionSignatures["heal_python"]], "heal_python")
             # display response
             config.print(config.divider)
             if config.developer:
@@ -223,8 +357,6 @@ class SharedUtil:
 
     @staticmethod
     def getDynamicTokens(messages, functionSignatures=None):
-        if not tiktokenImported:
-            return config.chatGPTApiMaxTokens
         if functionSignatures is None:
             functionTokens = 0
         else:
@@ -270,6 +402,8 @@ class SharedUtil:
                 "gpt-3.5-turbo-0613",
                 "gpt-3.5-turbo-16k",
                 "gpt-3.5-turbo-16k-0613",
+                "gpt-4-turbo-preview",
+                "gpt-4-0125-preview",
                 "gpt-4-1106-preview",
                 "gpt-4-0314",
                 "gpt-4-32k-0314",
@@ -339,7 +473,7 @@ Acess the risk level of this Python code:
                 model=config.chatGPTApiModel,
                 messages=[{"role": "user", "content" : userInput}],
                 n=1,
-                temperature=temperature if temperature is not None else config.chatGPTApiTemperature,
+                temperature=temperature if temperature is not None else config.llmTemperature,
                 max_tokens=config.chatGPTApiMaxTokens,
             )
             return completion.choices[0].message.content
@@ -366,9 +500,12 @@ Acess the risk level of this Python code:
 
     @staticmethod
     def get_wan_ip():
-        response = requests.get('https://api.ipify.org?format=json')
-        data = response.json()
-        return data['ip']
+        try:
+            response = requests.get('https://api.ipify.org?format=json', timeout=5)
+            data = response.json()
+            return data['ip']
+        except:
+            return ""
 
     @staticmethod
     def get_local_ip():
@@ -383,11 +520,50 @@ Acess the risk level of this Python code:
 
     @staticmethod
     def getDayOfWeek():
-        now = pendulum.now() 
-        return now.format('dddd')
+        if config.isTermux:
+            return ""
+        else:
+            now = pendulum.now() 
+            return now.format('dddd')
 
     @staticmethod
-    def getDeviceInfo():
+    def getWeather(latlng=""):
+        # get current weather information
+        # Reference: https://openweathermap.org/api/one-call-3
+
+        if not config.openweathermapApi:
+            return None
+
+        # latitude, longitude
+        if not latlng:
+            latlng = geocoder.ip('me').latlng
+
+        try:
+            latitude, longitude = latlng
+            # Build the URL for the weather API
+            url = f"https://api.openweathermap.org/data/2.5/weather?lat={latitude}&lon={longitude}&appid={config.openweathermapApi}&units=metric"
+            # Make the request to the API
+            response = requests.get(url)
+            # Parse the JSON response
+            data = json.loads(response.text)
+            # Get the current weather condition
+            weather_condition = data["weather"][0]["description"]
+            # Get the current temperature in Celsius
+            temperature_celsius = data["main"]["temp"]
+
+            # Convert the temperature to Fahrenheit
+            #temperature_fahrenheit = (temperature_celsius * 9/5) + 32
+
+            # Print the weather condition and temperature
+            #print(f"The current weather condition is {weather_condition}.")
+            #print(f"The current temperature is {temperature_fahrenheit} degrees Fahrenheit.")
+            return temperature_celsius, weather_condition
+        except:
+            SharedUtil.showErrors()
+            return None
+
+    @staticmethod
+    def getDeviceInfo(includeIp=False):
         g = geocoder.ip('me')
         if hasattr(config, "thisPlatform"):
             thisPlatform = config.thisPlatform
@@ -395,9 +571,19 @@ Acess the risk level of this Python code:
             thisPlatform = platform.system()
             if thisPlatform == "Darwin":
                 thisPlatform = "macOS"
-        wan_ip = SharedUtil.get_wan_ip()
-        local_ip = SharedUtil.get_local_ip()
-        dayOfWeek = SharedUtil.getDayOfWeek()
+        if config.includeIpInSystemMessageTemp or includeIp or (config.includeIpInSystemMessage and config.includeIpInSystemMessageTemp):
+            wan_ip = SharedUtil.get_wan_ip()
+            local_ip = SharedUtil.get_local_ip()
+            ipInfo = f'''Wan ip: {wan_ip}
+Local ip: {local_ip}
+'''
+        else:
+            ipInfo = ""
+        if config.isTermux:
+            dayOfWeek = ""
+        else:
+            dayOfWeek = SharedUtil.getDayOfWeek()
+            dayOfWeek = f"Current day of the week: {dayOfWeek}"
         return f"""Operating system: {thisPlatform}
 Version: {platform.version()}
 Machine: {platform.machine()}
@@ -409,10 +595,8 @@ Python version: {platform.python_version()}
 Python implementation: {platform.python_implementation()}
 Current directory: {os.getcwd()}
 Current time: {str(datetime.datetime.now())}
-Current day of the week: {dayOfWeek}
-Wan ip: {wan_ip}
-Local ip: {local_ip}
-Latitude & longitude: {g.latlng}
+{dayOfWeek}
+{ipInfo}Latitude & longitude: {g.latlng}
 Country: {g.country}
 State: {g.state}
 City: {g.city}"""
@@ -433,12 +617,7 @@ City: {g.city}"""
 
     @staticmethod
     def isPackageInstalled(package):
-        whichCommand = "where.exe" if platform.system() == "Windows" else "which"
-        try:
-            isInstalled, *_ = subprocess.Popen("{0} {1}".format(whichCommand, package), shell=True, stdout=subprocess.PIPE).communicate()
-            return True if isInstalled else False
-        except:
-            return False
+        return True if shutil.which(package.split(" ", 1)[0]) else False
 
     @staticmethod
     def getCliOutput(cli):
@@ -469,6 +648,16 @@ City: {g.city}"""
             response += f"\n# Error:\n{error}"
         return response
 
+    # Function to convert HTML to Markdown
+    @staticmethod
+    def convert_html_to_markdown(html_string):
+        # Create an instance of the HTML2Text converter
+        converter = html2text.HTML2Text()
+        # Convert the HTML string to Markdown
+        markdown_string = converter.handle(html_string)
+        # Return the Markdown string
+        return markdown_string
+
     @staticmethod
     def openURL(url):
         config.stopSpinning()
@@ -479,18 +668,18 @@ City: {g.city}"""
             webbrowser.open(url)
 
     @staticmethod
-    def getPreferredDir():
-        preferredDir = os.path.join(os.path.expanduser('~'), config.letMeDoItName.split()[0].lower())
+    def getStorageDir():
+        storageDir = os.path.join(os.path.expanduser('~'), config.letMeDoItName.split()[0].lower())
         try:
-            Path(preferredDir).mkdir(parents=True, exist_ok=True)
+            Path(storageDir).mkdir(parents=True, exist_ok=True)
         except:
             pass
-        return preferredDir if os.path.isdir(preferredDir) else ""
+        return storageDir if os.path.isdir(storageDir) else ""
 
     @staticmethod
     def getFiles():
-        preferredDir = SharedUtil.getPreferredDir()
-        if config.startupdirectory and not os.path.isdir(config.startupdirectory):
-            config.startupdirectory = ""
-        preferredDir = preferredDir if preferredDir else os.path.join(config.letMeDoItAIFolder, "files")
-        return config.startupdirectory if config.startupdirectory else preferredDir
\ No newline at end of file
+        storageDir = SharedUtil.getStorageDir()
+        if config.storagedirectory and not os.path.isdir(config.storagedirectory):
+            config.storagedirectory = ""
+        storageDir = storageDir if storageDir else os.path.join(config.letMeDoItAIFolder, "files")
+        return config.storagedirectory if config.storagedirectory else storageDir
\ No newline at end of file
